@@ -40,8 +40,17 @@ APPDIRS = appdirs.AppDirs('openstack', 'OpenStack', multipath='/etc')
 CONFIG_HOME = APPDIRS.user_config_dir
 CACHE_PATH = APPDIRS.user_cache_dir
 
-UNIX_CONFIG_HOME = os.path.join(
-    os.path.expanduser(os.path.join('~', '.config')), 'openstack')
+# snaps do set $HOME to something like
+# /home/$USER/snap/openstackclients/$SNAP_VERSION
+# the real home (usually /home/$USERNAME) is stored in $SNAP_REAL_HOME
+# see https://snapcraft.io/docs/environment-variables
+SNAP_REAL_HOME = os.getenv('SNAP_REAL_HOME')
+if SNAP_REAL_HOME:
+    UNIX_CONFIG_HOME = os.path.join(os.path.join(SNAP_REAL_HOME, '.config'),
+                                    'openstack')
+else:
+    UNIX_CONFIG_HOME = os.path.join(
+        os.path.expanduser(os.path.join('~', '.config')), 'openstack')
 UNIX_SITE_CONFIG_HOME = '/etc/openstack'
 
 SITE_CONFIG_HOME = APPDIRS.site_config_dir
@@ -70,6 +79,7 @@ VENDOR_FILES = [
 ]
 
 BOOL_KEYS = ('insecure', 'cache')
+CSV_KEYS = ('auth_methods',)
 
 FORMAT_EXCLUSIONS = frozenset(['password'])
 
@@ -128,7 +138,7 @@ def _fix_argv(argv):
                 options=','.join(overlap)))
 
 
-class OpenStackConfig(object):
+class OpenStackConfig:
 
     # These two attribute are to allow os-client-config to plumb in its
     # local versions for backwards compat.
@@ -151,9 +161,19 @@ class OpenStackConfig(object):
         self._load_envvars = load_envvars
 
         if load_yaml_config:
-            self._config_files = config_files or CONFIG_FILES
-            self._secure_files = secure_files or SECURE_FILES
-            self._vendor_files = vendor_files or VENDOR_FILES
+            # "if config_files" is not sufficient to process empty list
+            if config_files is not None:
+                self._config_files = config_files
+            else:
+                self._config_files = CONFIG_FILES
+            if secure_files is not None:
+                self._secure_files = secure_files
+            else:
+                self._secure_files = SECURE_FILES
+            if vendor_files is not None:
+                self._vendor_files = vendor_files
+            else:
+                self._vendor_files = VENDOR_FILES
         else:
             self._config_files = []
             self._secure_files = []
@@ -250,6 +270,7 @@ class OpenStackConfig(object):
                 clouds=dict(defaults=dict(self.defaults)))
             self.default_cloud = 'defaults'
 
+        self._cache_auth = False
         self._cache_expiration_time = 0
         self._cache_path = CACHE_PATH
         self._cache_class = 'dogpile.cache.null'
@@ -259,6 +280,9 @@ class OpenStackConfig(object):
         if 'cache' in self.cloud_config:
             cache_settings = _util.normalize_keys(self.cloud_config['cache'])
 
+            self._cache_auth = get_boolean(
+                cache_settings.get('auth', self._cache_auth))
+
             # expiration_time used to be 'max_age' but the dogpile setting
             # is expiration_time. Support max_age for backwards compat.
             self._cache_expiration_time = cache_settings.get(
@@ -289,25 +313,28 @@ class OpenStackConfig(object):
 
             influxdb_cfg = metrics_config.get('influxdb', {})
             # Parse InfluxDB configuration
-            if influxdb_config:
-                influxdb_cfg.update(influxdb_config)
-            if influxdb_cfg:
-                config = {}
-                if 'use_udp' in influxdb_cfg:
-                    use_udp = influxdb_cfg['use_udp']
-                    if isinstance(use_udp, str):
-                        use_udp = use_udp.lower() in ('true', 'yes', '1')
-                    elif not isinstance(use_udp, bool):
-                        use_udp = False
-                        self.log.warning('InfluxDB.use_udp value type is not '
-                                         'supported. Use one of '
-                                         '[true|false|yes|no|1|0]')
-                    config['use_udp'] = use_udp
-                for key in ['host', 'port', 'username', 'password', 'database',
-                            'measurement', 'timeout']:
-                    if key in influxdb_cfg:
-                        config[key] = influxdb_cfg[key]
-                self._influxdb_config = config
+            if not influxdb_config:
+                influxdb_config = influxdb_cfg
+            else:
+                influxdb_config.update(influxdb_cfg)
+
+        if influxdb_config:
+            config = {}
+            if 'use_udp' in influxdb_config:
+                use_udp = influxdb_config['use_udp']
+                if isinstance(use_udp, str):
+                    use_udp = use_udp.lower() in ('true', 'yes', '1')
+                elif not isinstance(use_udp, bool):
+                    use_udp = False
+                    self.log.warning('InfluxDB.use_udp value type is not '
+                                     'supported. Use one of '
+                                     '[true|false|yes|no|1|0]')
+                config['use_udp'] = use_udp
+            for key in ['host', 'port', 'username', 'password', 'database',
+                        'measurement', 'timeout']:
+                if key in influxdb_config:
+                    config[key] = influxdb_config[key]
+            self._influxdb_config = config
 
         if load_envvars:
             statsd_host = statsd_host or os.environ.get('STATSD_HOST')
@@ -906,7 +933,25 @@ class OpenStackConfig(object):
             # That it does not exist in keystoneauth is irrelvant- it not
             # doing what they want causes them sorrow.
             config['auth_type'] = 'admin_token'
-        return loading.get_plugin_loader(config['auth_type'])
+
+        loader = loading.get_plugin_loader(config['auth_type'])
+
+        # As the name would suggest, v3multifactor uses multiple factors for
+        # authentication. As a result, we need to register the configuration
+        # options for each required auth method. Normally, this is handled by
+        # the 'MultiFactor.load_from_options' method but there doesn't appear
+        # to be a way to "register" the auth methods without actually loading
+        # the plugin. As a result, if we encounter this auth type then we need
+        # to do this registration of extra options manually.
+        # FIXME(stephenfin): We need to provide a mechanism to extend the
+        # options in keystoneauth1.loading._plugins.identity.v3.MultiAuth
+        # without calling 'load_from_options'.
+        if config['auth_type'] == 'v3multifactor':
+            # We use '.get' since we can't be sure this key is set yet -
+            # validation happens later, in _validate_auth
+            loader._methods = config.get('auth_methods')
+
+        return loader
 
     def _validate_auth(self, config, loader):
         # May throw a keystoneauth1.exceptions.NoMatchingPlugin
@@ -1016,6 +1061,12 @@ class OpenStackConfig(object):
                 or ('token' in config and config['token'])):
             config.setdefault('token', config.pop('auth_token', None))
 
+        # Infer passcode if it was given separately
+        # This is generally absolutely impractical to require setting passcode
+        # in the clouds.yaml
+        if 'auth' in config and 'passcode' in config:
+            config['auth']['passcode'] = config.pop('passcode', None)
+
         # These backwards compat values are only set via argparse. If it's
         # there, it's because it was passed in explicitly, and should win
         config = self._fix_backwards_api_timeout(config)
@@ -1033,6 +1084,11 @@ class OpenStackConfig(object):
                 if type(config[key]) is not bool:
                     config[key] = get_boolean(config[key])
 
+        for key in CSV_KEYS:
+            if key in config:
+                if isinstance(config[key], str):
+                    config[key] = config[key].split(',')
+
         # TODO(mordred): Special casing auth_url here. We should
         #                come back to this betterer later so that it's
         #                more generalized
@@ -1119,6 +1175,21 @@ class OpenStackConfig(object):
         if not prefer_ipv6:
             force_ipv4 = True
 
+        # Override global metrics config with more specific per-cloud
+        # details.
+        metrics_config = config.get('metrics', {})
+        statsd_config = metrics_config.get('statsd', {})
+        statsd_host = statsd_config.get('host') or self._statsd_host
+        statsd_port = statsd_config.get('port') or self._statsd_port
+        statsd_prefix = statsd_config.get('prefix') or self._statsd_prefix
+        influxdb_config = metrics_config.get('influxdb', {})
+        if influxdb_config:
+            merged_influxdb = copy.deepcopy(self._influxdb_config)
+            merged_influxdb.update(influxdb_config)
+            influxdb_config = merged_influxdb
+        else:
+            influxdb_config = self._influxdb_config
+
         if cloud is None:
             cloud_name = ''
         else:
@@ -1134,16 +1205,17 @@ class OpenStackConfig(object):
             session_constructor=self._session_constructor,
             app_name=self._app_name,
             app_version=self._app_version,
+            cache_auth=self._cache_auth,
             cache_expiration_time=self._cache_expiration_time,
             cache_expirations=self._cache_expirations,
             cache_path=self._cache_path,
             cache_class=self._cache_class,
             cache_arguments=self._cache_arguments,
             password_callback=self._pw_callback,
-            statsd_host=self._statsd_host,
-            statsd_port=self._statsd_port,
-            statsd_prefix=self._statsd_prefix,
-            influxdb_config=self._influxdb_config,
+            statsd_host=statsd_host,
+            statsd_port=statsd_port,
+            statsd_prefix=statsd_prefix,
+            influxdb_config=influxdb_config,
         )
     # TODO(mordred) Backwards compat for OSC transition
     get_one_cloud = get_one
@@ -1239,6 +1311,7 @@ class OpenStackConfig(object):
             force_ipv4=force_ipv4,
             auth_plugin=auth_plugin,
             openstack_config=self,
+            cache_auth=self._cache_auth,
             cache_expiration_time=self._cache_expiration_time,
             cache_expirations=self._cache_expirations,
             cache_path=self._cache_path,
